# Importing necessary modules
import serial
import sys
import time
from roboticstoolbox import models
import numpy as np
import math

class AL5DControl:
    def __init__(self):
        
        # al5d model
        self.al5d = models.URDF.AL5D_mdw()

        self.robot_connected = False
        # search for port, initialise in home pose
        devices = ['/dev/ttyUSB0','/dev/ttyUSB1','COM1','COM2','COM3','COM4','COM5','COM6','COM7','COM8','COM9']
        
        # Starting the connection with the SSC-32U controller
        for device in devices:
            try:
                self.ssc32 = serial.Serial(device, 9600, timeout=1.0)
                print("Device connected on ", device)
                # Initializing the position of the robot
                self.robot_connected = True
                self.ssc32.write(b'#0P1500S100#1P1500S200#2P1500S300#3P1500S350#4P1500S100#5P1500S150\r')
            except serial.SerialException:
                print("No device connected on ", device)

        if (not self.robot_connected):
            print("Robot could not connect to any ports, restart kernel and try again")

    def collision_check(self, q0, q1, q2, q3, q4, gripper):
        q0r = q0*np.pi/180
        q1r = q1*np.pi/180
        q2r = q2*np.pi/180
        q3r = q3*np.pi/180
        q4r = q4*np.pi/180
        elbow = self.al5d.fkine([q0r,q1r,q2r,q3r,q4r], end='elbow')
        wrist = self.al5d.fkine([q0r,q1r,q2r,q3r,q4r], end='wrist')
        grip = self.al5d.fkine([q0r,q1r,q2r,q3r,q4r], end='gripper')
        tool = self.al5d.fkine([q0r,q1r,q2r,q3r,q4r], end='tool')
        
        if any(np.array([tool.t[2], grip.t[2], elbow.t[2], wrist.t[2]]) < -0.01):
            return True
        else:
            return False

    def robot_control(self,q0=0,q1=0,q2=0,q3=0,q4=0,gripper=0,dq0=85,dq1=85,dq2=85,dq3=85,dq4=85): #Method for controlling the joints coordinates and velocities
        if (self.robot_connected):
            if self.collision_check(q0,q1,q2,q3,q4,gripper):
                print("Robot collision with the table")
            else:
                print("No collision detected")
                pwq0 = str(round((q0+90)*11.1111 + 500)) # Scaling of the joint signal from (-90,90) to (500,2500)
                pwq1 = str(round((q1+90)*11.1111 + 500))
                pwq2 = str(round((q2+90)*11.1111 + 500))
                pwq3 = str(round((q3+90)*11.1111 + 500))
                pwq4 = str(round((q4+90)*11.1111 + 500))
                pwq5 = str(round((gripper+90)*11.1111 + 500))

                pwdq0 = str(round(dq0*11.1111))
                pwdq1 = str(round(dq1*11.1111))
                pwdq2 = str(round(dq2*11.1111))
                pwdq3 = str(round(dq3*11.1111))
                pwdq4 = str(round(dq4*11.1111))
                pwdq5 = str(1500)

                # Concatenating the desired control signals in a long string
                output="#0P"+pwq0+"S"+pwdq0+"#1P"+pwq1+"S"+pwdq1+"#2P"+pwq2+"S"+pwdq2+"#3P"+pwq3+"S"+pwdq3+"#4P"+pwq4+"S"+pwdq4+"#5P"+pwq5+"S"+pwdq5+"\r"
                # Formatting string
                output = output.encode('utf-8')  # Converting to bytes literals
                self.ssc32.write(output)  # sending serial data to the SSC-32 board


    def inst_AL5D_velocities(self, q0=0, q1=0, q2=0, q3=0, q4=0, joint='new pose'):
        if (self.robot_connected):
            np.set_printoptions(formatter={'float': '{: 0.4f}'.format})
            # scaling the angle inputs
            pwq0 = round((q0+90)*11.1111 + 500) # Scaling of the joint signal from (-90,90) to (500,2500)
            pwq1 = round((q1+90)*11.1111 + 500)
            pwq2 = round((q2+90)*11.1111 + 500)
            pwq3 = round((q3+90)*11.1111 + 500)
            pwq4 = round((q4+90)*11.1111 + 500)
            pwq5 = 1500

            # constructing the jacobian for a specific configuration
            rob = models.DH.AL5D_mdw()
            q   = np.array([q0,q1,q2,q3,q4])
            j   = rob.jacob0(q)

            # initialisation of position and velocity for servos
            pwq  = [pwq0, pwq1, pwq2, pwq3, pwq4, pwq5]
            pwdq = np.array([100, 100, 100, 100, 100])

            # checking which joint to jog
            if (joint == 'new pose'):
                i=-1
                output ="#0P"+str(pwq[0])+"S100"+"#1P"+str(pwq[1])+"S100"+"#2P"+str(pwq[2])+"S100"+"#3P"+str(pwq[3])+"S100"+"#4P"+str(pwq[4])+"S100"+"#5P"+str(pwq[5])+"S150"+"\r"
                output = output.encode('utf-8')
                self.ssc32.write(output)
                print("New pose achieved, ready for joints movements")

            if (joint == 'q0'):
                i=0
            if (joint == 'q1'):
                i=1
            if (joint == 'q2'):
                i=2
            if (joint == 'q3'):
                i=3
            if (joint == 'q4'):
                i=4

            # jogging chosen joint forward and backward
            if (i>=0):
                for count in range(10):
                    if (count%2 == 0):
                        s = 1
                    else:
                        s = -1

                    pwq[i] = pwq[i] + s*500
                    output ="#0P"+str(pwq[0])+"S"+str(pwdq[0])+"#1P"+str(pwq[1])+"S"+str(pwdq[1])+"#2P"+str(pwq[2])+"S"+str(pwdq[2])+"#3P"+str(pwq[3])+"S"+str(pwdq[3])+"#4P"+str(pwq[4])+"S"+str(pwdq[4])+"#5P1500"+"S150"+"\r"
                    output = output.encode('utf-8')
                    time.sleep(0.8)
                    self.ssc32.write(output)

            print("The jacobian in this configuration is \n")
            print(j)
            print("\nInstantaneous velocities of the end-effector for joint velocities "+str(pwdq*1.5707/(11.1111*180))+"[rad/s] in this configuration are \n")
            vel_rad = (pwdq*1.5707)/(11.1111*180)
            print(j@np.transpose(vel_rad))
